
#include <stdio.h>

#if USE_PROTO
#include <stdlib.h>
#else
extern char *           malloc();
extern char *           realloc();
extern void             free();
#define size_t          unsigned int
#endif

#include "chunk.h"

struct S_ChunkHdr;
typedef struct S_ChunkHdr           CHUNKHDR;
typedef CHUNKHDR *                  pCHUNKHDR;
struct S_ChunkHdr
{
    pCHUNKHDR           pNxtChunk;
    size_t              nOfs;
    char                acData[1];
};


#if USE_PROTO
pALLOCCHUNK ChunkInit(size_t size)
#else
pALLOCCHUNK ChunkInit(size)
size_t size;
#endif
{
    auto     pALLOCCHUNK        p;
    auto     pCHUNKHDR          pHdr;

    if (size < 32)
        return(NULL);

    p = (pALLOCCHUNK) malloc(sizeof(ALLOCCHUNK));
    if (p)
    {
        p->ChunkSize  = size;
        pHdr = (pCHUNKHDR) malloc(size + sizeof(CHUNKHDR));
        if (pHdr)
        {
            p->pChunkList   = (char *) pHdr;
            p->pChunkActive = (char *) pHdr;
            pHdr->pNxtChunk = NULL;
            pHdr->nOfs = 0;
        }
        else
        {
            free(p);
            p = NULL;
        }
    }

    return(p);
}


#if USE_PROTO
void ChunkTerm(pALLOCCHUNK p)
#else
void ChunkTerm(p)
pALLOCCHUNK p;
#endif
{
    auto     pCHUNKHDR          p1;

    if (p)
    {
        p1 = (pCHUNKHDR) p->pChunkList;
        while (p1)
        {
            auto     pCHUNKHDR          p2 = p1->pNxtChunk;

            free(p1);
            p1 = p2;
        }

        free(p);
    }

    return;
}


#if USE_PROTO
char *ChunkAlloc(pALLOCCHUNK p, size_t size)
#else
char *ChunkAlloc(p, size)
pALLOCCHUNK p;
size_t size;
#endif
{
    auto     size_t         sizeFree;
    auto     char *         pRet;
    auto     pCHUNKHDR      pHdr = (pCHUNKHDR) p->pChunkActive;

    if (size & 3)
        size = (size & ~3) + 4;

    if (size > p->ChunkSize)
        return(NULL);

    if (pHdr->nOfs > p->ChunkSize)
        sizeFree = 0;
    else
        sizeFree = p->ChunkSize - pHdr->nOfs;

    if (sizeFree < size)
    {
        pHdr->pNxtChunk = (pCHUNKHDR) malloc(p->ChunkSize + sizeof(CHUNKHDR));
        if (NULL == pHdr->pNxtChunk)
            return(NULL);

        pHdr = pHdr->pNxtChunk;
        pHdr->pNxtChunk = NULL;
        pHdr->nOfs = 0;
        p->pChunkActive = (char *) pHdr;
    }

    pRet = pHdr->acData + pHdr->nOfs;
    pHdr->nOfs += size;
    return(pRet);
}


#if USE_PROTO
void ChunkFree(pALLOCCHUNK p, char *pMem)
#else
void ChunkFree(p, pMem)
pALLOCCHUNK p;
char *pMem;
#endif
{
    auto     pCHUNKHDR          pHdr = (pCHUNKHDR) p->pChunkList;

    while (pHdr)
    {
        auto     char *             pCh = pHdr->acData;

        /* -----------------2/25/96 9:55AM-------------------
         * 16 bit versions of the Microsoft compiler only
         * compare the offset portion of far pointers.  as a
         * result we need to case the pointers to be huge so
         * the entire 32 bit values will be compared.
         * --------------------------------------------------*/
#if defined(_MSC_VER) && (_MSC_VER < 900)
        if ((char huge *) pMem >= (char huge *) pCh
                               &&
            (char huge *) pMem < (char huge *) pCh + pHdr->nOfs)
#else
        if (pMem >= pCh && pMem < pCh + pHdr->nOfs)
#endif
        {
            auto     pCHUNKHDR          pNxt = pHdr->pNxtChunk;

            pHdr->nOfs = pMem - pCh;
            pHdr->pNxtChunk = NULL;
            p->pChunkActive = (char *) pHdr;
            pHdr = pNxt;
            break;
        }

        pHdr = pHdr->pNxtChunk;
    }

    while (pHdr)
    {
        auto     pCHUNKHDR      pNxt = pHdr->pNxtChunk;

        free(pHdr);
        pHdr = pNxt;
    }

    return;
}
