/*
 *  WINREF - Locate unresolved Windows dynalinks
 *
 *          Copyright (c) 1994, Bob Withers
 *
 *          bwit@pobox.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include <direct.h>
#include <io.h>

#include "winref.h"
#include "chunk.h"

#define MAXDIM(a)                       (sizeof(a) / sizeof(a[0]))
#define MAX_MODNAME_LEN                 10

struct S_ModEntry;
typedef struct S_ModEntry               MODENTRY;
typedef MODENTRY *                      pMODENTRY;

struct S_ExpEntry;
typedef struct S_ExpEntry               EXPENTRY;
typedef EXPENTRY *                      pEXPENTRY;

struct S_ImpEntry;
typedef struct S_ImpEntry               IMPENTRY;
typedef IMPENTRY *                      pIMPENTRY;

struct S_ModNameNode;
typedef struct S_ModNameNode            MODNAMENODE;
typedef MODNAMENODE *                   pMODNAMENODE;

struct S_ExpEntry
{
    pEXPENTRY       pNxtExp;
    int             nNumRefs;
    ushort          usOrdinal;
    char *          pszName;
};

union U_ModPtr
{
    char *          pszModName;
    pMODENTRY       pModEntry;
};
typedef union U_ModPtr                  MODPTR;
typedef MODPTR *                        pMODPTR;

struct S_ImpEntry
{
    pIMPENTRY       pNxtImp;
    pEXPENTRY       pExp;
    MODPTR          Mod;
    ushort          usOrdinal;
    char *          pszName;
};

struct S_ModEntry
{
    pMODENTRY       pNxtMod;
    int             bPresent;
    char            szModName[MAX_MODNAME_LEN];
    char *          pszModPath;
    uchar           ucModType;          /* from NE ucExeType        */
    uchar           ucOtherFlags;       /* from NE ucOtherFlags     */
    int             nNumExp;
    pEXPENTRY       pExp;
    int             nNumImp;
    pIMPENTRY       pImp;
};

struct S_ModNameNode
{
    pMODNAMENODE    pNxtNode;
    char            szModName[1];
};


struct S_SysIniVars
{
    char *          pszSysIniKey;
    char *          pszModName;
    char *          pszNewMod;
};
typedef struct S_SysIniVars             SYSINIVARS;
typedef SYSINIVARS *                    pSYSINIVARS;


static int          AccessModule(char *pszModName);
static void         AddExpEntry(pMODENTRY pMod, pEXPENTRY pExp);
static void         AddImpEntry(pMODENTRY pMod, pIMPENTRY pImp);
static char *       AddModName(char *pszModName);
static pEXPENTRY    CreateExpEntry(ushort usOrdinal, int nNameLen,
                                    char *pszName);
static void         DumpModules(void);
static void         DumpUnresolved(void);
static void         ExtractModName(char *pszPath, char *pszModName,
                                    int nMaxLen);
static char *       FindModName(char *pszModName);
static pMODENTRY    FindModule(char *pszModName);
static void         LoadAllModules(char *pszName);
static bool         LoadModule(char *pszName);
static char *       LoadModuleInfo(FILE *pf, pOLDEXEHDR pOldHdr,
                                    pNEWEXEHDR pNewHdr, char *pszName,
                                    char *pszPathName);
static char *       LocateModule(char *pszName, char *pszPathName,
                                    int nBufLen);
static void         Message(bool bTerminate, char *pszFmt, ...);
static void         ProcessArgs(int *argc, char **argv);
static char *       ProcessExports(FILE *pf, pOLDEXEHDR pOldHdr,
                                    pNEWEXEHDR pNewHdr, pMODENTRY pMod);
static char *       ProcessImports(FILE *pf, pOLDEXEHDR pOldHdr,
                                    pNEWEXEHDR pNewHdr, pMODENTRY pMod);
static char *       ProcSegRelocs(FILE *pf, pSEGTABENTRY pSegTab,
                                    pMODENTRY pMod, ushort *pusModRef,
                                    char *pImpNam, ushort usSectorSize);
static void         ProcessSystemIni(void);
static char *       ReadNewExeHdr(FILE *pf, pOLDEXEHDR pOldHdr,
                                    pNEWEXEHDR pNewHdr);
static char *       ReadOldExeHdr(FILE *pf, pOLDEXEHDR pOldHdr);
static void         ResolveDynalinks(pMODENTRY pMod);
static void         ResolveImport(pIMPENTRY pImp);
static void         Usage(int bTerminate);


static pMODENTRY        pModList = NULL;
static pMODNAMENODE     pModNameList = NULL;

static SYSINIVARS       SysIniVars[] =
            {
                /* entry for the Windows kernel must be first!  */
                { "WinKernel",      "KERNEL.DLL",       "KRNL386.EXE"   },
                { "display.drv",    "DISPLAY.DLL",      NULL            },
                { "network.drv",    "NETWORK.DLL",      NULL            },
                { "mouse.drv",      "MOUSE.DLL",        NULL            },
                { "sound.drv",      "SOUND.DLL",        NULL            },
                { "comm.drv",       "COMM.DLL",         NULL            },
                { "keyboard.drv",   "KEYBOARD.DLL",     NULL            },
                { "system.drv",     "SYSTEM.DLL",       NULL            }
            };

static int              bVerboseMode = 0;
static char             szVersion[]  = "0.6 (beta)";
static pALLOCCHUNK      pChunk;
static char *           pszWinDir    = "C:\\WINDOWS";
static char             szErrReading[] = "\nError reading %s [%s]";
static char             szOutOfMem[] = "Out of memory";
static char             szWrk[512];


int main(int argc, char **argv)
{
    auto     char           szModName[FILENAME_MAX];

    ProcessArgs(&argc, argv);
    if (argc < 2)
        Usage(1);

    Usage(0);
    pChunk = ChunkInit(16000);
    if (NULL == pChunk)
        Message(TRUE, szOutOfMem);

    ProcessSystemIni();
    strcpy(szModName, argv[1]);
    LoadAllModules(strupr(szModName));
    if (bVerboseMode)
        DumpModules();
    else
        DumpUnresolved();

    ChunkTerm(pChunk);
    return(0);
}


static void ProcessArgs(int *argc, char **argv)
{
    register int            i = 1;
    auto     char *         p;

    while (i < *argc)
    {
        if ('/' == argv[i][0] || '-' == argv[i][0])
        {
            switch (argv[i][1])
            {
                case 'v':
                case 'V':
                    bVerboseMode = 1;
                    break;

                case 'k':
                case 'K':
                    p = argv[i] + 2;
                    while (' ' == *p)
                        ++p;

                    if ('\0' == *p)
                    {
                        if (--(*argc) > i)
                        {
                            memmove(&argv[i], &argv[i + 1],
                                    (*argc - i) * sizeof(argv[i]));
                            p = argv[i];
                        }
                        else
                        {
                            Usage(1);
                        }
                    }

                    SysIniVars[0].pszNewMod = strupr(p);
                    break;

                case 'w':
                case 'W':
                    p = argv[i] + 2;
                    while (' ' == *p)
                        ++p;

                    if ('\0' == *p)
                    {
                        if (--(*argc) > i)
                        {
                            memmove(&argv[i], &argv[i + 1],
                                    (*argc - i) * sizeof(argv[i]));
                            p = argv[i];
                        }
                        else
                        {
                            Usage(1);
                        }
                    }

                    pszWinDir = strupr(p);
                    break;

                default:
                    Message(FALSE, "\nInvalid switch %s ignored", argv[i]);
                    break;
            }

            if (--(*argc) > i)
            {
                memmove(&argv[i], &argv[i + 1],
                        (*argc - i) * sizeof(argv[i]));
            }
        }
        else
        {
            ++i;
        }
    }

    return;
}


static void LoadAllModules(char *pszName)
{
    register pIMPENTRY          pImp;
    auto     int                nModsLoaded;
    auto     pMODENTRY          pMod;
    auto     char               szModName[MAX_MODNAME_LEN + 4];

    if (LoadModule(pszName))
        Message(FALSE, "\n\nChecking dynalink references of %s", pszName);
    else
        Message(TRUE, "\n\nUnable to locate module %s\n", pszName);

    while (TRUE)
    {
        nModsLoaded = 0;
        pMod = pModList;
        while (pMod)
        {
            /*-----------------08-31-94 07:50am-----------------
             after a module is loaded scan for unloaded module
             references and load them.  this was being done
             much more elegantly using recursion but the cost
             in stack space was excessive.  as a result this
             iterative approach will have to suffice.
            --------------------------------------------------*/
            pImp = pMod->pImp;
            while (pImp)
            {
                if (NULL == pImp->pExp)
                {
                    if (NULL == FindModule(pImp->Mod.pszModName))
                    {
                        ++nModsLoaded;
                        strcpy(szModName, pImp->Mod.pszModName);
                        strcat(szModName, ".DLL");
                        LoadModule(szModName);
                    }
                }

                pImp = pImp->pNxtImp;
            }

            pMod = pMod->pNxtMod;
        }

        if (0 == nModsLoaded)
            break;
    }

    /*-----------------08-31-94 07:52am-----------------
     now that we should have all the referenced modules
     loaded, go through and resolve all the imported
     dynalinks.
    --------------------------------------------------*/
    pMod = pModList;
    while (pMod)
    {
        ResolveDynalinks(pMod);
        pMod = pMod->pNxtMod;
    }

    return;
}


static pMODENTRY FindModule(char *pszModName)
{
    auto     pMODENTRY          pMod = pModList;

    while (pMod)
    {
        if (0 == strcmp(pMod->szModName, pszModName))
            break;

        pMod = pMod->pNxtMod;
    }

    return(pMod);
}


static bool LoadModule(char *pszName)
{
    auto     FILE *             pf;
    auto     char *             pszMsg;
    auto     OLDEXEHDR          OldHdr;
    auto     NEWEXEHDR          NewHdr;
    auto     char               szPathName[FILENAME_MAX];

    LocateModule(pszName, szPathName, sizeof(szPathName));
    pf = fopen(szPathName, "rb");
    if (NULL == pf)
    {
        pszMsg = LoadModuleInfo(NULL, NULL, NULL, pszName, szPathName);
        if (pszMsg)
            Message(TRUE, pszMsg);

        return(FALSE);
    }

    pszMsg = ReadOldExeHdr(pf, &OldHdr);
    if (NULL == pszMsg)
        pszMsg = ReadNewExeHdr(pf, &OldHdr, &NewHdr);

    if (NULL == pszMsg)
        pszMsg = LoadModuleInfo(pf, &OldHdr, &NewHdr, pszName, szPathName);

    if (pszMsg)
        Message(TRUE, szErrReading, szPathName, pszMsg);

    fclose(pf);
    return(TRUE);
}


static char *ReadOldExeHdr(FILE *pf, pOLDEXEHDR pOldHdr)
{
    if (1 != fread(pOldHdr, sizeof(*pOldHdr), 1, pf))
        return("Unable to read old EXE header");

    if (OLDEXE_MAGIC != pOldHdr->usMagic)
        return("Invalid executable file");

    if (0L == pOldHdr->lNewExeOfs)
        return("This is an old style DOS executable");

    return(NULL);
}


static char *ReadNewExeHdr(FILE *pf, pOLDEXEHDR pOldHdr, pNEWEXEHDR pNewHdr)
{
    if (0 != fseek(pf, pOldHdr->lNewExeOfs, SEEK_SET))
        return("Unable to position to new EXE header");

    if (1 != fread(pNewHdr, sizeof(*pNewHdr), 1, pf))
        return("Unable to read new EXE header");

    if (NEWEXE_MAGIC != pNewHdr->usMagic)
        return("Invalid new EXE header magic number");

    return(NULL);
}


static char *LoadModuleInfo(FILE *pf, pOLDEXEHDR pOldHdr,
                            pNEWEXEHDR pNewHdr, char *pszName,
                            char *pszPathName)
{
    auto     pMODENTRY          pMod;
    auto     char *             pszMsg;

    pMod = (pMODENTRY) ChunkAlloc(pChunk, sizeof(*pMod));
    if (NULL == pMod)
        return("Unable to alloc MOD entry");

    pMod->pszModPath = ChunkAlloc(pChunk, strlen(pszPathName) + 1);
    if (NULL == pMod->pszModPath)
        return("Unable to alloc MOD path");

    pMod->pNxtMod = NULL;
    ExtractModName(pszName, pMod->szModName, sizeof(pMod->szModName));
    strcpy(pMod->pszModPath, pszPathName);
    if (pf)
    {
        pMod->bPresent = 1;
        pMod->ucModType = pNewHdr->ucExeType;
        pMod->ucOtherFlags = pNewHdr->ucOtherFlags;
        pszMsg = ProcessExports(pf, pOldHdr, pNewHdr, pMod);
        if (NULL == pszMsg)
            pszMsg = ProcessImports(pf, pOldHdr, pNewHdr, pMod);

        if (pszMsg)
            return(pszMsg);
    }
    else
    {
        pMod->bPresent = 0;
        pMod->ucModType = 0;
        pMod->ucOtherFlags = 0;
        pMod->nNumExp = 0;
        pMod->pExp = NULL;
        pMod->nNumImp = 0;
        pMod->pImp = NULL;
    }

    /*-----------------08-27-94 08:05am-----------------
     add the new module to the end of the module list.
    --------------------------------------------------*/
    if (pModList)
    {
        auto     pMODENTRY          p = pModList;

        while (p->pNxtMod)
            p = p->pNxtMod;

        p->pNxtMod = pMod;
    }
    else
    {
        pModList = pMod;
    }

    return(NULL);
}


static char *ProcessExports(FILE *pf, pOLDEXEHDR pOldHdr,
                            pNEWEXEHDR pNewHdr, pMODENTRY pMod)
{
    auto     long               lOfs;
    auto     uchar              ucLen;
    auto     size_t             len;
    auto     pEXPENTRY          pExp;
    auto     ushort             usOrdinal;
    auto     int                bFirstEntry;

    pMod->nNumExp = 0;
    pMod->pExp = NULL;

    /*-----------------08-27-94 08:58am-----------------
     process the resident names table.
    --------------------------------------------------*/
    bFirstEntry = 1;
    lOfs = pOldHdr->lNewExeOfs + (long) pNewHdr->usResNameTab;
    if (0 != fseek(pf, lOfs, SEEK_SET))
        return("Unable to position to resident name table");

    if (1 != fread(&ucLen, sizeof(ucLen), 1, pf))
        return("Unable to read resident name length");

    while (ucLen)
    {
        len = (size_t) ucLen;
        if (1 != fread(szWrk, len + sizeof(ushort), 1, pf))
            return("Unable to read resident name");

        /*-----------------08-27-94 08:49am-----------------
         the first entry in the resident names table is the
         name of this module, we just ignore it.
        --------------------------------------------------*/
        if (bFirstEntry)
        {
            bFirstEntry = 0;
        }
        else
        {
            pExp = CreateExpEntry(*((ushort *) (szWrk + len)), len, szWrk);
            if (NULL == pExp)
                return("Unable to alloc export entry");

            AddExpEntry(pMod, pExp);
        }

        if (1 != fread(&ucLen, sizeof(ucLen), 1, pf))
            return("Unable to read resident name length");
    }

    /*-----------------08-27-94 08:59am-----------------
     now process the non-resident names table.
    --------------------------------------------------*/
    bFirstEntry = 1;
    if (0 != fseek(pf, pNewHdr->lNonResNameTab, SEEK_SET))
        return("Unable to position to non-resident name table");

    if (1 != fread(&ucLen, sizeof(ucLen), 1, pf))
        return("Unable to read non-resident name length");

    while (ucLen)
    {
        len = (size_t) ucLen;
        if (1 != fread(szWrk, len + sizeof(ushort), 1, pf))
            return("Unable to read non-resident name");

        /*-----------------08-27-94 08:49am-----------------
         the first entry in the non-resident names table is the
         description of this module, we just ignore it.
        --------------------------------------------------*/
        if (bFirstEntry)
        {
            bFirstEntry = 0;
        }
        else
        {
            pExp = CreateExpEntry(*((ushort *) (szWrk + len)), len, szWrk);
            if (NULL == pExp)
                return("Unable to alloc export entry");

            AddExpEntry(pMod, pExp);
        }

        if (1 != fread(&ucLen, sizeof(ucLen), 1, pf))
            return("Unable to read non-resident name length");
    }

    /* -----------------2/23/96 6:49AM-------------------
     * finally process the Entry Table to pick up references
     * that are exported by ordinal only
     * --------------------------------------------------*/
    lOfs = pOldHdr->lNewExeOfs + (long) pNewHdr->usEntryTab;
    if (0 != fseek(pf, lOfs, SEEK_SET))
        return("Unable to position to entry table");

    if (1 != fread(&ucLen, sizeof(ucLen), 1, pf))
        return("Unable to read entry table bundle length");

    usOrdinal = 0;
    while (ucLen)
    {
        auto     uchar              ucSegInd;

        if (1 != fread(&ucSegInd, sizeof(ucSegInd), 1, pf))
            return("Unable to read entry table bundle segind");

        switch (ucSegInd)
        {
            case 0:
                usOrdinal += ucLen;
                break;

            case 0xff:
                while (ucLen--)
                {
                    auto     MOVESEGENTRY       MoveSeg;

                    if (1 != fread(&MoveSeg, MOVESEGENTRY_SIZE, 1, pf))
                        return("Unable to read entry table moveable");

                    ++usOrdinal;
                    if (MoveSeg.ucFlags & MOVESEGENTRY_EXPORTED)
                    {
                        pExp = CreateExpEntry(usOrdinal, 0, NULL);
                        if (NULL == pExp)
                            return("Unable to alloc export entry");

                        AddExpEntry(pMod, pExp);
                    }
                }

                break;

            default:
                while (ucLen--)
                {
                    auto     FIXEDSEGENTRY      FixedSeg;

                    if (1 != fread(&FixedSeg, FIXEDSEGENTRY_SIZE, 1, pf))
                        return("Unable to read entry table fixed");

                    ++usOrdinal;
                    if (FixedSeg.ucFlags & FIXEDSEGENTRY_EXPORTED)
                    {
                        pExp = CreateExpEntry(usOrdinal, 0, NULL);
                        if (NULL == pExp)
                            return("Unable to alloc export entry");

                        AddExpEntry(pMod, pExp);
                    }
                }

                break;
        }

        if (1 != fread(&ucLen, sizeof(ucLen), 1, pf))
            return("Unable to read entry table bundle length");
    }

    return(NULL);
}


static pEXPENTRY CreateExpEntry(ushort usOrdinal, int nNameLen, char *pszName)
{
    auto     pEXPENTRY          pExp;

    pExp = (pEXPENTRY) ChunkAlloc(pChunk, sizeof(*pExp));
    if (NULL == pExp)
        return(NULL);

    pExp->pNxtExp = NULL;
    pExp->nNumRefs = 0;
    pExp->usOrdinal = usOrdinal;
    pExp->pszName = ChunkAlloc(pChunk, nNameLen + 1);
    if (NULL == pExp->pszName)
        return(NULL);

    if (nNameLen > 0)
        memcpy(pExp->pszName, pszName, nNameLen);

    pExp->pszName[nNameLen] = '\0';
    return(pExp);
}


static void AddExpEntry(pMODENTRY pMod, pEXPENTRY pExp)
{
    auto     pEXPENTRY          p1;
    auto     pEXPENTRY          p2;

    ++pMod->nNumExp;
    if (NULL == pMod->pExp)
    {
        pExp->pNxtExp = NULL;
        pMod->pExp = pExp;
        return;
    }

    p1 = p2 = pMod->pExp;
    while (p2)
    {
        if (pExp->usOrdinal == p2->usOrdinal)
        {
            ChunkFree(pChunk, (char *) pExp);
            return;
        }

        if (pExp->usOrdinal < p2->usOrdinal)
        {
            if (p1 == p2)
            {
                pExp->pNxtExp = p1;
                pMod->pExp = pExp;
            }
            else
            {
                pExp->pNxtExp = p1->pNxtExp;
                p1->pNxtExp = pExp;
            }

            break;
        }

        if (NULL == p2->pNxtExp)
        {
            pExp->pNxtExp = p2->pNxtExp;
            p2->pNxtExp = pExp;
            break;
        }

        p1 = p2;
        p2 = p2->pNxtExp;
    }

    return;
}


static char *ProcessImports(FILE *pf, pOLDEXEHDR pOldHdr,
                            pNEWEXEHDR pNewHdr, pMODENTRY pMod)
{
    auto     ushort         i;
    auto     ushort         usSectorSize;
    auto     size_t         len;
    auto     long           lOfs;
    auto     ushort *       pusModRef;
    auto     char *         pImpNames;
    auto     pSEGTABENTRY   pSegTab;
    auto     char *         pszMsg;

    pMod->nNumImp = 0;
    pMod->pImp = NULL;
    if (0 == pNewHdr->usNoModRef)
        return(NULL);

    len = pNewHdr->usNoModRef * sizeof(*pusModRef);
    pusModRef = (ushort *) malloc(len);
    if (NULL == pusModRef)
        return(szOutOfMem);

    lOfs = pOldHdr->lNewExeOfs + (long) pNewHdr->usModRefTab;
    if (0 != fseek(pf, lOfs, SEEK_SET))
        return("Unable to position to module ref table");

    if (1 != fread(pusModRef, len, 1, pf))
        return("Unable to read module ref table");

    len = pNewHdr->usEntryTab - pNewHdr->usImpNameTab;
    pImpNames = malloc(len);
    if (NULL == pImpNames)
        return(szOutOfMem);

    lOfs = pOldHdr->lNewExeOfs + (long) pNewHdr->usImpNameTab;
    if (0 != fseek(pf, lOfs, SEEK_SET))
        return("Unable to position to imported names table");

    if (1 != fread(pImpNames, len, 1, pf))
        return("Unable to read imported names table");

    len = pNewHdr->usNoSegs * sizeof(*pSegTab);
    pSegTab = (pSEGTABENTRY) malloc(len);
    if (NULL == pSegTab)
        return(szOutOfMem);

    lOfs = pOldHdr->lNewExeOfs + (long) pNewHdr->usSegTab;
    if (0 != fseek(pf, lOfs, SEEK_SET))
        return("Unable to position to segment table");

    if (1 != fread(pSegTab, len, 1, pf))
        return("Unable to read segment table");

    usSectorSize = 1U << pNewHdr->usAlign;
    for (i = 0; i < pNewHdr->usNoSegs; ++i)
    {
        pszMsg = ProcSegRelocs(pf, pSegTab + i, pMod,
                                pusModRef, pImpNames, usSectorSize);
        if (pszMsg)
            return(pszMsg);
    }

    free(pusModRef);
    free(pImpNames);
    free(pSegTab);
    return(NULL);
}


static void AddImpEntry(pMODENTRY pMod, pIMPENTRY pImp)
{
    auto     pIMPENTRY          p1;
    auto     pIMPENTRY          p2;
    auto     char *             pszNewName;
    auto     int                i;

    ++pMod->nNumImp;
    if (NULL == pMod->pImp)
    {
        pMod->pImp = pImp;
        pImp->pNxtImp = NULL;
        return;
    }

    if (pImp->pExp)
        pszNewName = pImp->Mod.pModEntry->szModName;
    else
        pszNewName = pImp->Mod.pszModName;

    p1 = p2 = pMod->pImp;
    while (p2)
    {
        auto     char *         pszName;

        if (p2->pExp)
            pszName = p2->Mod.pModEntry->szModName;
        else
            pszName = p2->Mod.pszModName;

        i = strcmp(pszNewName, pszName);
        if (0 == i)
        {
            i = (int) ((long) pImp->usOrdinal - (long) p2->usOrdinal);
            if (0 == i)
            {
                /*-----------------08-31-94 07:34am-----------------
                    if module name, ordinal, and entry name are equal
                    then this is a duplicate import relocation, we
                    only need to save one of them.
                --------------------------------------------------*/
                if (pImp->usOrdinal)
                {
                    --pMod->nNumImp;
                    return;
                }

                i = strcmp(pImp->pszName, p2->pszName);
                if (0 == i)
                {
                    --pMod->nNumImp;
                    return;
                }
            }
        }

        if (i <= 0)
        {
            if (p1 == p2)
            {
                pImp->pNxtImp = pMod->pImp;
                pMod->pImp = pImp;
            }
            else
            {
                pImp->pNxtImp = p1->pNxtImp;
                p1->pNxtImp = pImp;
            }

            break;
        }

        if (NULL == p2->pNxtImp)
        {
            pImp->pNxtImp = NULL;
            p2->pNxtImp = pImp;
            break;
        }

        p1 = p2;
        p2 = p2->pNxtImp;
    }

    return;
}


static char *ProcSegRelocs(FILE *pf, pSEGTABENTRY pSegTab, pMODENTRY pMod,
                            ushort *pusModRef, char *pImpNames,
                            ushort usSectorSize)
{
    register ushort         i;
    auto     char *         p;
    auto     ushort         usNoRelocs;
    auto     SEGRELOCENTRY  SegReloc;
    auto     long           lOfs;
    auto     pIMPENTRY      pImp;

    if (!(pSegTab->usSectorOfs > 0
                &&
        pSegTab->usFlags & SEGFLG_NEEDS_RELOC))
    {
        return(NULL);
    }

    lOfs = (long) usSectorSize * (long) pSegTab->usSectorOfs;
    if (0 == pSegTab->usSegLen)
        lOfs += 65535L;
    else
        lOfs += (long) pSegTab->usSegLen;

    if (0 != fseek(pf, lOfs, SEEK_SET))
        return("Unable to position to segment reloc");

    if (1 != fread(&usNoRelocs, sizeof(usNoRelocs), 1, pf))
        return("Unable to read segment reloc count");

    for (i = 0; i < usNoRelocs; ++i)
    {
        if (1 != fread(&SegReloc, sizeof(SegReloc), 1, pf))
            return("Unable to read segment reloc entry");

        if (RELOCADDR_PTR_32 == SegReloc.usRelocAddrType
                            &&
            (RELOCTYP_IMP_ORD == SegReloc.usRelocType
                            ||
             RELOCTYP_IMP_NAM == SegReloc.usRelocType))
        {
            pImp = (pIMPENTRY) ChunkAlloc(pChunk, sizeof(*pImp));
            if (NULL == pImp)
                return(szOutOfMem);

            pImp->pExp = NULL;
            p = pImpNames + pusModRef[SegReloc.U.ImpOrd.usModIdx - 1];
            memcpy(szWrk, p + 1, *p);
            szWrk[*p] = '\0';
            pImp->Mod.pszModName = AddModName(szWrk);
            if (RELOCTYP_IMP_NAM == SegReloc.usRelocType)
            {
                pImp->usOrdinal = 0;
                p = pImpNames + SegReloc.U.ImpNam.usImpName;
                memcpy(szWrk, p + 1, *p);
                szWrk[*p] = '\0';
                pImp->pszName = AddModName(szWrk);
            }
            else
            {
                pImp->usOrdinal = SegReloc.U.ImpOrd.usOrdinal;
                pImp->pszName = NULL;
            }

            AddImpEntry(pMod, pImp);
        }
    }

    return(NULL);
}


static void ResolveDynalinks(pMODENTRY pMod)
{
    register pIMPENTRY      pImp = pMod->pImp;

    while (pImp)
    {
        ResolveImport(pImp);
        pImp = pImp->pNxtImp;
    }

    return;
}


static void ResolveImport(pIMPENTRY pImp)
{
    auto     pMODENTRY      pMod;
    auto     pEXPENTRY      pExp;

    if (pImp->pExp)
        return;

    pMod = FindModule(pImp->Mod.pszModName);
    if (pMod)
    {
        pExp = pMod->pExp;
        if (pImp->usOrdinal > 0)
        {
            while (pExp)
            {
                if (pImp->usOrdinal == pExp->usOrdinal)
                {
                    pImp->Mod.pModEntry = pMod;
                    pImp->pExp = pExp;
                    ++pExp->nNumRefs;
                    return;
                }

                pExp = pExp->pNxtExp;
            }
        }
        else
        {
            while (pExp)
            {
                if (0 == strcmp(pImp->pszName, pExp->pszName))
                {
                    pImp->Mod.pModEntry = pMod;
                    pImp->pExp = pExp;
                    ++pExp->nNumRefs;
                    return;
                }

                pExp = pExp->pNxtExp;
            }
        }
    }

    return;
}

static char *LocateModule(char *pszName, char *pszPathName, int nBufLen)
{
    auto     int            i;
    auto     char *         p;
    auto     char *         pszPath = NULL;

    for (i = 0; i < MAXDIM(SysIniVars); ++i)
    {
        if (0 == stricmp(SysIniVars[i].pszModName, pszName))
        {
            if (SysIniVars[i].pszNewMod)
                pszName = SysIniVars[i].pszNewMod;

            break;
        }
    }

    /*-----------------08-21-94 07:14am-----------------
     if one of the following is true then assume that the
     caller has specified a fully qualified path.
    --------------------------------------------------*/
    if (':' == pszName[1] || strpbrk(pszName, "/\\"))
    {
        strcpy(pszPathName, pszName);
        return(pszPathName);
    }

    /*-----------------08-21-94 11:06am-----------------
     check to see if file is in current directory.
    --------------------------------------------------*/
    if (0 == AccessModule(pszName))
    {
        getcwd(pszPathName, nBufLen);
        strcat(pszPathName, "\\");
        strcat(pszPathName, pszName);
        return(pszPathName);
    }

    p = getenv("PATH");
    if (p)
    {
        pszPath = malloc(strlen(p) + 1);
        if (pszPath)
            strcpy(pszPath, p);
    }

    if (pszPath)
    {
        p = strtok(pszPath, ";");
        while (p)
        {
            strcpy(pszPathName, p);
            i = (int) strlen(pszPathName) - 1;
            if (i >= 0)
            {
                if (!('\\' == pszPathName[i] || '/' == pszPathName[i]))
                    strcat(pszPathName, "\\");
            }

            strcat(pszPathName, pszName);
            if (0 == AccessModule(pszPathName))
            {
                free(pszPath);
                return(pszPathName);
            }

            p = strtok(NULL, ";");
        }
    }

    if (pszPath)
        free(pszPath);

    /*-----------------08-29-94 02:29pm-----------------
     if a Windows directory is provided look there.
    --------------------------------------------------*/
    if (pszWinDir)
    {
        i = strlen(pszWinDir);
        strcpy(pszPathName, pszWinDir);
        if (i > 0
                &&
            !('\\' == pszPathName[i - 1] || '/' == pszPathName[i - 1]))
        {
            strcat(pszPathName, "\\");
        }

        strcat(pszPathName, pszName);
        if (0 == AccessModule(pszPathName))
            return(pszPathName);

        strcpy(pszPathName, pszWinDir);
        if (i > 0
                &&
            !('\\' == pszPathName[i - 1] || '/' == pszPathName[i - 1]))
        {
            strcat(pszPathName, "\\");
        }

        strcat(pszPathName, "SYSTEM\\");
        strcat(pszPathName, pszName);
        if (0 == AccessModule(pszPathName))
            return(pszPathName);
    }

    strcpy(pszPathName, pszName);
    return(pszPathName);
}


static int AccessModule(char *pszModName)
{
    auto     int            i;
    auto     char *         p;
    static   char *         pszExt[] =
                            {
                                ".DLL",
                                ".EXE",
                                ".DRV",
                                ".FON"
                            };

    if (0 == access(pszModName, 0))
        return(0);

    strcpy(szWrk, pszModName);
    p = strrchr(szWrk, '.');
    if (NULL == p)
        p = szWrk + strlen(szWrk);

    for (i = 0; i < MAXDIM(pszExt); ++i)
    {
        strcpy(p, pszExt[i]);
        if (0 == access(szWrk, 0))
        {
            strcpy(pszModName, szWrk);
            return(0);
        }
    }

    return(-1);
}


static void DumpUnresolved(void)
{
    auto     pMODENTRY      pMod;
    auto     pIMPENTRY      pImp;
    auto     int            nUnresolved = 0;

    pMod = pModList;
    while (pMod)
    {
        pImp = pMod->pImp;
        while (pImp)
        {
            if (NULL == pImp->pExp)
            {
                if (1 == ++nUnresolved)
                    Message(FALSE, "\n");

                if (pImp->usOrdinal)
                    sprintf(szWrk, "%u", (unsigned) pImp->usOrdinal);
                else
                {
                    if (pImp->pszName)
                        strcpy(szWrk, pImp->pszName);
                    else
                        strcpy(szWrk, "??unknown??");
                }

                Message(FALSE, "\n%s [%s] refers to unresolved %s.%s",
                        pMod->szModName, pMod->pszModPath,
                        pImp->Mod.pszModName, szWrk);
            }

            pImp = pImp->pNxtImp;
        }

        pMod = pMod->pNxtMod;
    }

    Message(FALSE, "\n\n    %d unresolved dynalink%s",
            nUnresolved, 1 == nUnresolved ? "" : "s");
    return;
}


static void DumpModules(void)
{
    auto     pMODENTRY      pMod;
    auto     pEXPENTRY      pExp;
    auto     pIMPENTRY      pImp;

    pMod = pModList;
    while (pMod)
    {
        Message(FALSE, "\n\nDynaLink Information for %s [%s]",
                    pMod->szModName, pMod->pszModPath);
        if (pMod->bPresent)
        {
            if (pMod->pExp)
            {
                Message(FALSE, "\n\n  %d Export%s:", pMod->nNumExp,
                        1 == pMod->nNumExp ? "" : "s");
                Message(FALSE, "\n     Ord RefCt   Name");
                pExp = pMod->pExp;
                while (pExp)
                {
                    Message(FALSE, "\n   %5u  %3d   %s",
                                (unsigned int) pExp->usOrdinal,
                                pExp->nNumRefs, pExp->pszName);

                    pExp = pExp->pNxtExp;
                }
            }

            if (pMod->pImp)
            {
                auto     char *         pszName;

                Message(FALSE, "\n\n  %d Import%s:", pMod->nNumImp,
                        1 == pMod->nNumImp ? "" : "s");
                pImp = pMod->pImp;
                while (pImp)
                {
                    if (pImp->pExp)
                    {
                        if (pImp->usOrdinal)
                        {
                            sprintf(szWrk, "\n    %s.%u ",
                                    pImp->Mod.pModEntry->szModName,
                                    (unsigned int) pImp->usOrdinal);
                            pszName = pImp->pExp->pszName;
                        }
                        else
                        {
                            sprintf(szWrk, "\n    %s.%s ",
                                    pImp->Mod.pModEntry->szModName,
                                    pImp->pszName);
                            pszName = NULL;
                        }
                    }
                    else
                    {
                        pszName = "** Unresolved **";
                        if (pImp->usOrdinal)
                        {
                            sprintf(szWrk, "\n    %s.%u ",
                                    pImp->Mod.pszModName,
                                    (unsigned int) pImp->usOrdinal);
                        }
                        else
                        {
                            sprintf(szWrk, "\n    %s.%s ",
                                    pImp->Mod.pszModName,
                                    pImp->pszName);
                        }
                    }

                    if (pszName)
                    {
                        while (strlen(szWrk) < 35)
                            strcat(szWrk, " ");

                        strcat(szWrk, pszName);
                    }

                    Message(FALSE, szWrk);
                    pImp = pImp->pNxtImp;
                }
            }
        }
        else
        {
            Message(FALSE, "\n\n    **** This module was not found ****");
        }

        pMod = pMod->pNxtMod;
    }

    return;
}


static void ExtractModName(char *pszPath, char *pszModName, int nMaxLen)
{
    auto     char *         p;
    auto     char *         p2;
    auto     int            i;

    p = strrchr(pszPath, '\\');
    if (NULL == p)
        p = strrchr(pszPath, '/');

    if (NULL == p)
        p = pszPath;
    else
        ++p;


    p2 = strrchr(p, '.');
    if (NULL == p2)
        p2 = p + strlen(p);

    i = p2 - p;
    if (i >= nMaxLen)
        i = nMaxLen - 1;

    memcpy(pszModName, p, i);
    pszModName[i] = '\0';
    strupr(pszModName);
    if (0 == strcmp(pszModName, "KRNL386"))
        strcpy(pszModName, "KERNEL");

    return;
}


static char *FindModName(char *pszModName)
{
    auto     pMODNAMENODE           p = pModNameList;

    while (p)
    {
        if (0 == strcmp(pszModName, p->szModName))
            return(p->szModName);

        p = p->pNxtNode;
    }

    return(NULL);
}


static char *AddModName(char *pszModName)
{
    auto     pMODNAMENODE           pNode;
    auto     char *                 p;
    auto     size_t                 len;

    p = FindModName(pszModName);
    if (p)
        return(p);

    len = sizeof(*pNode) + strlen(pszModName);
    pNode = (pMODNAMENODE) ChunkAlloc(pChunk, len);
    if (NULL == pNode)
        Message(TRUE, "Unable to alloc a ModNameNode");

    strcpy(pNode->szModName, pszModName);
    pNode->pNxtNode = pModNameList;
    pModNameList = pNode;
    return(pNode->szModName);
}


static void ProcessSystemIni(void)
{
    register int            i;
    register char *         p;
    auto     char *         p2;
    auto     char *         p3;
    auto     FILE *         pfSysIni;
    auto     char           c;
    auto     int            bFound = 0;

    strcpy(szWrk, pszWinDir);
    c = szWrk[strlen(szWrk) - 1];
    if (!('\\' == c || '/' == c))
        strcat(szWrk, "\\");

    strcat(szWrk, "SYSTEM.INI");
    pfSysIni = fopen(szWrk, "r");
    if (NULL == pfSysIni)
        return;

    /*-----------------08-30-94 01:50pm-----------------
     try to find the [boot] section of system.ini
    --------------------------------------------------*/
    while (fgets(szWrk, sizeof(szWrk), pfSysIni))
    {
        p = szWrk;
        while (isspace(*p))
            ++p;

        if ('[' == *p)
        {
            ++p;
            p2 = strchr(p, ']');
            if (p2)
            {
                *p2 = '\0';
                if (0 == stricmp(p, "boot"))
                {
                    bFound = 1;
                    break;
                }
            }
        }
    }

    if (bFound)
    {
        while (fgets(szWrk, sizeof(szWrk), pfSysIni))
        {
            p = szWrk;
            while (isspace(*p))
                ++p;

            if ('[' == *p)
                break;

            if ('=' == *p)
                continue;

            p2 = strchr(p, '=');
            if (p2)
            {
                p3 = p2;
                while (isspace(*(p3 - 1)))
                    --p3;

                *p3 = '\0';
                for (i = 0; i < MAXDIM(SysIniVars); ++i)
                {
                    auto     int            len;

                    if (SysIniVars[i].pszNewMod)
                        continue;

                    if (0 == stricmp(p, SysIniVars[i].pszSysIniKey))
                    {
                        ++p2;
                        while (isspace(*p2))
                            ++p2;

                        p3 = strchr(p2, '\n');
                        if (p3)
                            *p3 = '\0';

                        len = strlen(p2);
                        if (len > 0)
                        {
                            SysIniVars[i].pszNewMod = malloc(len + 1);
                            if (SysIniVars[i].pszNewMod)
                                strcpy(SysIniVars[i].pszNewMod, strupr(p2));
                        }

                        break;
                    }
                }
            }
        }
    }

    fclose(pfSysIni);
    return;
}


static void Usage(int bTerminate)
{
    Message(FALSE,
            "\nWINREF - Locate Unresolved MS Windows DynaLinks - V%s",
            szVersion);
    Message(FALSE, "\n         Copyright (c) 1994 - Bob Withers");
    Message(FALSE, "\n         Permission granted to freely copy");
    if (bTerminate)
    {
        Message(FALSE, "\n\n  Usage: WINREF [-v] [-w<path>] [-k<modname>] "
                       "filename");
        Message(FALSE, "\n             -v           Verbose mode");
        Message(FALSE, "\n             -w           Path to Windows DIR "
                       "(Default C:\\WINDOWS)");
        Message(FALSE, "\n             -k           Name of Windows Kernel "
                       "(Default KNRL386.EXE)");
        Message(TRUE,  "\n             filename     Name of file to check");
    }

    return;
}


static void Message(bool bTerminate, char *pszFmt, ...)
{
    auto     va_list            ap;

    va_start(ap, pszFmt);
    vprintf(pszFmt, ap);
    va_end(ap);
    if (bTerminate)
        exit(1);

    return;
}
