
#ifndef WINREF_H
#define WINREF_H

#include "config.h"

#pragma pack(1)

typedef unsigned char           uchar;
typedef unsigned short          ushort;
typedef unsigned long           ulong;
typedef signed int              bool;

#ifndef TRUE
#define TRUE                    1
#define FALSE                   0
#endif

struct S_OldExeHdr
{
    ushort          usMagic;            /* Magic number MZ              */
        #define OLDEXE_MAGIC            0x5A4D
    ushort          usBytesLp;          /* Bytes on last page of file   */
    ushort          usPages;            /* Pages in file                */
    ushort          usRelocs;           /* Relocations                  */
    ushort          usHdrSize;          /* Size of header in paragraphs */
    ushort          usMinAlloc;         /* Min extra paragraphs needed  */
    ushort          usMaxAlloc;         /* Max extra paragraphs needed  */
    ushort          usSS;               /* Initial (relative) SS value  */
    ushort          usSP;               /* Initial SP value             */
    ushort          usChkSum;           /* Checksum                     */
    ushort          usIP;               /* Initial IP value             */
    ushort          usCS;               /* Initial (relative) CS value  */
    ushort          usRelocTab;         /* File addr of reloc table     */
    ushort          usOvlNo;            /* Overlay number               */
    ushort          usReserved1[4];     /* Reserved words               */
    ushort          usOemId;            /* OEM identifier               */
    ushort          usOemInfo;          /* OEM info, OEM specific       */
    ushort          usReserved2[10];    /* Reserved words               */
    long            lNewExeOfs;         /* File address of new exe hdr  */
};
typedef struct S_OldExeHdr              OLDEXEHDR;
typedef OLDEXEHDR *                     pOLDEXEHDR;


struct S_NewExeHdr
{
    ushort          usMagic;            /* Magic number NE              */
        #define NEWEXE_MAGIC            0x454E
    uchar           ucVer;              /* Version number               */
    uchar           ucRevision;         /* Revision number              */
    ushort          usEntryTab;         /* Offset of Entry Table        */
    ushort          usEntryBytes;       /* Number bytes in Entry Table  */
    long            lChkSum;            /* Checksum of whole file       */
    ushort          usFlags;            /* Flag word                    */
    ushort          usAutoSegNo;        /* Auto data segment number     */
    ushort          usHeapAlloc;        /* Initial heap allocation      */
    ushort          usStackAlloc;       /* Initial stack allocation     */
    ulong           ulCSIP;             /* Initial CS:IP setting        */
    ulong           ulSSSP;             /* Initial SS:SP setting        */
    ushort          usNoSegs;           /* Count of file segments       */
    ushort          usNoModRef;         /* Entries in Module Ref Table  */
    ushort          usModRefSize;       /* Size non-resident name table */
    ushort          usSegTab;           /* Offset of Segment Table      */
    ushort          usResTab;           /* Offset of Resource Table     */
    ushort          usResNameTab;       /* Offset of resident name table*/
    ushort          usModRefTab;        /* Offset of Module Ref Table   */
    ushort          usImpNameTab;       /* Offset of Imported Names Tab */
    long            lNonResNameTab;     /* Offset of Non-res Names Table*/
    ushort          usMoveCnt;          /* Count of movable entries     */
    ushort          usAlign;            /* Segment alignment shift cnt  */
    ushort          usResCnt;           /* Count of resource entries    */
    uchar           ucExeType;          /* Target operating system      */
        #define NETYPE_UNKNOWN          0
        #define NETYPE_OS2              1
        #define NETYPE_WINDOWS          2
        #define NETYPE_DOS4             3
        #define NETYPE_WIN386           4
    uchar           ucOtherFlags;       /* Other .EXE flags             */
        #define NEFLG_WIN2_PROT_MODE    0x01
        #define NEFLG_WIN2_PROP_FONT    0x02
        #define NEFLG_FAST_LOAD_AREA    0x04
    ushort          usFastLoadArea;     /* Fast load area               */
    ushort          usFastLoadLen;      /* Length of fast load area     */
    ushort          usReserved;         /* Reserved                     */
    uchar           ucWinMinVer;        /* Expected Windows minor ver   */
    uchar           ucWinMajVer;        /* Expected Windows major ver   */
};
typedef struct S_NewExeHdr              NEWEXEHDR;
typedef NEWEXEHDR *                     pNEWEXEHDR;


struct S_SegTabEntry
{
    ushort          usSectorOfs;        /* Logical sector offset of seg */
    ushort          usSegLen;           /* Len of segment (0 = 64K)     */
    ushort          usFlags;            /* Segment flags                */
        #define SEGFLG_DATA_SEG         0x0001
        #define SEGFLG_MEM_ALLOCED      0x0002
        #define SEGFLG_SEG_LOADED       0x0004
        #define SEGFLG_SEG_MOVEABLE     0x0010
        #define SEGFLG_SEG_SHARABLE     0x0020
        #define SEGFLG_SEG_PRELOAD      0x0040
        #define SEGFLG_SEG_READONLY     0x0080      /* data segments    */
        #define SEGFLG_SEG_EXECONLY     0x0080      /* code segments    */
        #define SEGFLG_NEEDS_RELOC      0x0100
        #define SEGFLG_SEG_CONFORMING   0x0200
        #define SEGFLG_SEG_DISCARDABLE  0x1000
    ushort          usMinAlloc;         /* Min alloc size (0 = 64K)     */
};
typedef struct S_SegTabEntry            SEGTABENTRY;
typedef SEGTABENTRY *                   pSEGTABENTRY;


struct S_RelocFixIntRef
{
    uchar           ucSegNo;
    uchar           ucFiller;               /* set to zero          */
    ushort          usSegOfs;
};
typedef struct S_RelocFixIntRef         RELOCFIXINTREF;
typedef RELOCFIXINTREF *                pRELOCFIXINTREF;

struct S_RelocMovIntRef
{
    uchar           ucSegNo;                /* set to FF            */
    uchar           ucFiller;               /* set to zero          */
    ushort          usEntryOrd;             /* entry table ordinal  */
};
typedef struct S_RelocMovIntRef         RELOCMOVINTREF;
typedef RELOCMOVINTREF *                pRELOCMOVINTREF;

struct S_RelocImpOrd
{
    ushort          usModIdx;               /* index into mod ref table */
    ushort          usOrdinal;              /* function ordinal number  */
};
typedef struct S_RelocImpOrd            RELOCIMPORD;
typedef RELOCIMPORD *                   pRELOCIMPORD;

struct S_RelocImpNam
{
    ushort          usModIdx;               /* index into mod ref table */
    ushort          usImpName;              /* ofs into imported names  */
};
typedef struct S_RelocImpNam            RELOCIMPNAM;
typedef RELOCIMPNAM *                   pRELOCIMPNAM;

union U_SegRelocUnique
{
    RELOCFIXINTREF      FixIntRef;
    RELOCMOVINTREF      MovIntRef;
    RELOCIMPORD         ImpOrd;
    RELOCIMPNAM         ImpNam;
};
typedef union U_SegRelocUnique          SEGRELOCUNIQUE;
typedef SEGRELOCUNIQUE *                pSEGRELOCUNIQUE;

struct S_SegRelocEntry
{
    uchar           usRelocAddrType;
        #define RELOCADDR_LOW_BYTE      0
        #define RELOCADDR_SEL_16        2
        #define RELOCADDR_PTR_32        3
        #define RELOCADDR_OFS_16        5
        #define RELOCADDR_PTR_48        11
        #define RELOCADDR_OFS_32        13
    uchar           usRelocType;
        #define RELOCTYP_INT_REF        0       /* internal reference   */
        #define RELOCTYP_IMP_ORD        1       /* imported ordinal     */
        #define RELOCTYP_IMP_NAM        2       /* imported name        */
        #define RELOCTYP_OSFIXUP        3       /* OS fixup             */
    ushort          usRelocOfs;
    SEGRELOCUNIQUE  U;
};
typedef struct S_SegRelocEntry          SEGRELOCENTRY;
typedef SEGRELOCENTRY *                 pSEGRELOCENTRY;


#define FIXEDSEGENTRY_SIZE              3
struct S_FixedSegEntry
{
    uchar           ucFlags;
        #define FIXEDSEGENTRY_EXPORTED  1       /* The entry is exported    */
        #define FIXEDSEGENTRY_GLOBAL    2       /* uses global data segs    */
    ushort          usSegOfs;
};
typedef struct S_FixedSegEntry          FIXEDSEGENTRY;
typedef FIXEDSEGENTRY *                 pFIXEDSEGENTRY;


#define MOVESEGENTRY_SIZE               6
struct S_MoveSegEntry
{
    uchar           ucFlags;
        #define MOVESEGENTRY_EXPORTED   1       /* The entry is exported    */
        #define MOVESEGENTRY_GLOBAL     2       /* uses global data segs    */
    ushort          ucInt3F;                    /* contains Int 3fh         */
    uchar           ucSegment;                  /* Segment number           */
    ushort          usSegOfs;
};
typedef struct S_MoveSegEntry           MOVESEGENTRY;
typedef MOVESEGENTRY *                  pMOVESEGENTRY;

#pragma pack()

#endif      /* WINREF_H */
